#cython: boundscheck=False, wraparound=False, nonecheck=True, cdivision = True

import numpy as np
from libc.math cimport sin, cos, sqrt


def fe(double t, double T, int M, double [:] u):
    cdef int k
    cdef double y, w
    cdef double pi = np.pi
    y = u[0] / sqrt(T)
    for k in range(M):
        w = 2*(k+1)*pi/T
        y += (2/sqrt(T)) * (u[2*k+1]*sin(w*t) +
                            u[2*k+2]*cos(w*t))
    return y


def dfe(double t, double T, int M, double [:] u, int m):
    cdef int m2
    cdef double w
    cdef double pi = np.pi
    if(m == 0):
        return 1.0/sqrt(T)
    elif(m % 2 == 0):
        m2 = int(m/2)
        w = 2*m2*np.pi/T
        return (2/sqrt(T)) * cos(w*t)
    else:
        m2 = int((m+1)/2)
        w = 2*m2*np.pi/T
        return (2/sqrt(T)) * sin(w*t)


def gradfe(double t, double T, int M, double [:] u):
    cdef int m2
    cdef double w
    cdef double pi = np.pi
    cdef int m
    grad = np.zeros(2*M+1)
    for m in range(2*M+1):
        if(m == 0):
            #return 1.0/sqrt(T)
            grad[m] = 1.0/sqrt(T)
        elif(m % 2 == 0):
            m2 = int(m/2)
            w = 2*m2*np.pi/T
            #return (2/sqrt(T)) * cos(w*t)
            grad[m] = (2/sqrt(T)) * cos(w*t)
        else:
            m2 = int((m+1)/2)
            w = 2*m2*np.pi/T
            #return (2/sqrt(T)) * sin(w*t)
            grad[m] = (2/sqrt(T)) * sin(w*t)
    return grad


def fourierexpansion(t, double T, int nu, double [:] u):
    cdef int M, j
    M = int((nu-1)/2)
    cdef double [:] ts
    cdef double [:] yy
    if isinstance(t, float):
        return fe(t, T, M, u)
    else:
        ts = t
        yt = np.zeros(t.shape[0])
        yy = yt
        for j in range(t.shape[0]):
            yy[j] = fe(ts[j], T, M, u)
        return yt


def dfourierexpansion(t, double T, int nu, double [:] u, int m):
    cdef int M, j
    M = int((nu-1)/2)
    cdef double [:] ts
    cdef double [:] yy
    if isinstance(t, float):
        return dfe(t, T, M, u, m)
    else:
        ts = t
        yt = np.zeros(t.shape[0])
        yy = yt
        for j in range(t.shape[0]):
            yy[j] = dfe(ts[j], T, M, u, m)
        return yt


def gradfourierexpansion(t, double T, int nu, double [:] u):
    cdef int M, j
    M = int((nu-1)/2)
    cdef double [:] ts
    cdef double [:] yy
    if isinstance(t, float):
        #return dfe(t, T, M, u, m)
        return gradfe(t, T, M, u)
    else:
        ts = t
        #yt = np.zeros(t.shape[0])
        yt = np.zeros([t.shape[0], nu])
        #yy = yt
        for j in range(t.shape[0]):
            #yy[j] = dfe(ts[j], T, M, u, m)
            yt[j, :] = dfe(ts[j], T, M, u)
        return yt


cdef int kdelta(int i, int j):
    if i == j:
        return 1
    else:
        return 0


def floquetcomponent(int dim, int ntsteps, double[:] ws, double complex[:, :, :] lindw):
    cdef int fdim
    cdef int alpha
    cdef int beta
    cdef int m, n, i, j
    fdim = dim * ntsteps
    lindb = np.zeros([fdim, fdim], dtype = complex)
    cdef double complex[:, :] lindb_view = lindb
    for i in range(fdim):
        alpha = i // ntsteps
        n = i % ntsteps
        for j in range(fdim):
            beta = j // ntsteps
            m = j % ntsteps
            if n-m < 0:
                lindb_view[i, j] = lindw[alpha, beta, ntsteps + (n-m)] - 1j * ws[m] * kdelta(i, j)
            else:
                lindb_view[i, j] = lindw[alpha, beta, n-m] - 1j * ws[m] * kdelta(i, j)
    return lindb


def floquetcomponent2(int Nu, int dim, int ntsteps, double complex[:, :, :, :] dlindw):
    cdef int fdim
    cdef int alpha
    cdef int beta
    cdef int m, n, i, j
    fdim = dim * ntsteps
    dlindb = np.zeros([Nu, fdim, fdim], dtype = complex)
    cdef double complex[:, :, :] dlindb_view = dlindb
    for k in range(Nu):
        for i in range(fdim):
            alpha = i // ntsteps
            n = i % ntsteps
            for j in range(fdim):
                beta = j // ntsteps
                m = j % ntsteps
                if n-m < 0:
                    dlindb_view[k, i, j] = dlindw[k, alpha, beta, ntsteps +  (n-m)]
                else:
                    dlindb_view[k, i, j] = dlindw[k, alpha, beta, n-m]
    return dlindb



## Copyright 2019-present The qocttools developing team
##
## This file is part of qocttools.
##
## qocttools is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## qocttools is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with qocttools.  If not, see <https://www.gnu.org/licenses/>.

"""This module holds the class that holds the Hamiltonian.


"""

import numpy as np
import qutip as qt

class hamiltonian:
    """A class to hold and manipulate the Hamiltonian and dissipation terms.

    The Hamiltonian should contain a static part, a list of possible perturbations,
    and also a list of dissipative terms for the Lindblad equation (which is not,
    of course, a proper term of the Hamiltonian)

    There are two options for the user to pass the Hamiltonian to qocttools: as
    a list of Qobj objects, or as a Qobj-valued user-defined function.

    1. In the first case, the Hamiltonian has the form:

       .. math::
          H(t) = H_0 + \sum_{i=1}^N f_i(t) V_i

       The pulses :math:`f_i(t)` are not described in the hamiltonian object.
       The user must then pass H0 and a list of perturbations V (it may also
       be just one). In addition, for open systems, the hamiltonian object
       should also contain the Lindblad operators, which should be supplied
       by the user as a list of Qobj objects.

    2. In the second case, the user should supply Qobj-valued functions. In this
       case, the H0 function contains the full Hamiltonian, and depends on time
       and on the set of "pulses" :math:`f_i(t)`:

       .. math::
          H_0 = H_0(t, f_1(t), \dots, f_N(t))

       The user must then also supply, in the argument V, a list with the
       derivatives of :math:`H0` with respect to each :math:`f_i` argument:

       .. math::
          V_i(t, f_1(t), \dots, f_N(t)) = \\frac{\partial H_o}{\partial f_i}(t, f_1(t), \dots, f_N(t))

       The function H0 (and the functions in the V list) should have the following inteface:

       .. code-block:: python

          def H0(t, args):
              f = args["f"]
              # f[0], f[1], f[2], ..., f[N-1] are the time-dependent functions

       The arguments are a float t that means time, and a dictionary args with just one element
       with name "f", and whose value is a list of functions of time.


    Parameters
    ----------
    H0 : Qobj or Qobj-valued function
        The qutip Qobj instance holding the static part of the Hamiltonian.
    V : Qobj or list of Qobj or Qobj-valued function or list of Qobj-valued functions.
        One or various Qobj instances holding external perturbations.
    A : Qobj or list of Qobj, default = None
        One or various Qobj instances holding the jump operators defining
        the Lindbladian.
    g : still undocumnted, default = None
    gradg: still undocumented, default = None    

    Attributes
    ----------
    H0 : Qobj
        The qutip Qobj instance holding the static part of the Hamiltonian.
    V : list of Qobj
        One or various Qobj instances holding external perturbations.
    A : list of Qobj
        One or various Qobj instances holding the jump operators defining
        the Lindblad operator.

    """

    def __init__(self, H0, V, A = None, g = None, gradg = None):
        if not isinstance(H0, qt.qobj.Qobj):
            self.function = True
            self.H0 = H0
            if isinstance(V, list):
                self.V = []
                for j in range(len(V)):
                    (self.V).append(V[j])
            else:
                self.V = [V]
            self.dim = H0(0.0, args = { "f" : [lambda t : 0]*len(self.V) } ).dims[0][0]
        else:
            self.function = False
            self.dim = H0.dims[0][0]
            self.H0 = H0.copy()
            if isinstance(V, list):
                self.V = []
                for j in range(len(V)):
                    (self.V).append(V[j].copy())
            else:
                self.V = [V.copy()]

        self.A = A
        if g is not None:
            self.g = g
        if gradg is not None:
            self.gradg = gradg

    def has_dissipative_terms(self):
        """Returns True if the dissipative terms are not None

        Returns
        -------
        bool:
            True if the attribute holding the dissipative terms
            is not None, False otherwise.

        """
        return (self.A is not None)

    def gradg(self, f, j, l):
        if j == l:
            return 1
        else:
            return 0

    def g(self, f, j):
        return f[j]


def toliouville(h):
    """
    This function transforms a Hamiltonian, defined in Hilbert
    space, together with the set of Lindblad operators, to the corresponding
    Liouvillian in Liouville space. It is also returned as a hamiltonian
    object

    Parameters
    ----------
    h : hamiltonian
       The input Hamiltonian, to be transformed.

    Returns
    -------
    hamiltonian
       A hamiltonian object with the transformed Hamiltonian.

    """
    dim = h.dim**2
    #print("dim = {}".format(dim))
    v__ = []
    l0__ = qt.Qobj(qt.liouvillian(h.H0, h.A), dims = [[dim], [dim]])
    for j in range(len(h.V)):
        v__.append( qt.Qobj(qt.liouvillian(h.V[j]), dims = [[dim], [dim]]) )
    return hamiltonian(l0__, v__)




import numpy as np
import jax.numpy as jnp
import qutip as qt
import jax; jax.config.update('jax_platform_name', 'cpu')
from jax import grad
#import reciprocal.kpoints as kpoints
from reciprocal.kpoints import Kpointset
#import os

angstromtobohr = 1.8897261
eVtohartree = 0.036749322

#TNN parameters for MoS2 (GGA) (data from Table III)  
aa   = 3.190 * angstromtobohr 
e1   = 0.683 * eVtohartree
e2   = 1.707 * eVtohartree
t0   =-0.146 * eVtohartree
t1   =-0.114 * eVtohartree
t2   = 0.506 * eVtohartree
t11  = 0.085 * eVtohartree
t12  = 0.162 * eVtohartree
t22  = 0.073 * eVtohartree
r0   = 0.060 * eVtohartree
r1   =-0.236 * eVtohartree
r2   = 0.067 * eVtohartree
r11  = 0.016 * eVtohartree
r12  = 0.087 * eVtohartree
u0   =-0.038 * eVtohartree
u1   = 0.046 * eVtohartree
u2   = 0.001 * eVtohartree
u11  = 0.266 * eVtohartree
u12  =-0.176 * eVtohartree
u22  =-0.150 * eVtohartree

#TNN functions
def getab(kx, ky):
    # if (len(kgrid.shape)<2):
    #     a,b = (kgrid[0],kgrid[1])
    # else:
    #     a,b = (kgrid[:,0],kgrid[:,1])
    a,b = (aa * kx, aa * ky)
    return a*0.5,b*jnp.sqrt(3)*0.5


def V0(kx, ky):
    a,b = getab(kx, ky)
    v0 = e1 + 2*t0*(2*jnp.cos(a)  *jnp.cos(b)  + jnp.cos(2*a))\
            + 2*r0*(2*jnp.cos(3*a)*jnp.cos(b)  + jnp.cos(2*b))\
            + 2*u0*(2*jnp.cos(2*a)*jnp.cos(2*b)+ jnp.cos(4*a))
    return v0
V0x = grad(V0, argnums = 0)
V0y = grad(V0, argnums = 1)

def reV1(kx, ky):
    a,b = getab(kx, ky)
    rev1 = -2*jnp.sqrt(3)*t2*jnp.sin(a)*jnp.sin(b) + 2*(r1+r2)*jnp.sin(3*a)*jnp.sin(b) \
           -2*jnp.sqrt(3)*u2*jnp.sin(2*a)*jnp.sin(2*b)
    return rev1
reV1x = grad(reV1, argnums = 0)
reV1y = grad(reV1, argnums = 1)

def imV1(kx, ky):
    a,b = getab(kx, ky)
    imv1 = 2*t1*jnp.sin(a)*(2*jnp.cos(a)+jnp.cos(b)) + 2*(r1-r2)*jnp.sin(3*a)*jnp.cos(b) \
          +2*u1*jnp.sin(2*a)*(2*jnp.cos(2*a)+jnp.cos(2*b)) 
    return imv1
imV1x = grad(imV1, argnums = 0)
imV1y = grad(imV1, argnums = 1)
    
def reV2(kx, ky):
    a,b = getab(kx, ky)
    rev2 = 2*t2*(jnp.cos(2*a)-jnp.cos(a)*jnp.cos(b))\
          -2/jnp.sqrt(3)*(r1+r2)*(jnp.cos(3*a)*jnp.cos(b)-jnp.cos(2*b))\
          +2*u2*(jnp.cos(4*a)-jnp.cos(2*a)*jnp.cos(2*b))
    return rev2
reV2x = grad(reV2, argnums = 0)
reV2y = grad(reV2, argnums = 1)

def imV2(kx, ky):
    a,b = getab(kx, ky)
    imv2 = 2*jnp.sqrt(3)*t1*jnp.cos(a)*jnp.sin(b) \
          +2/jnp.sqrt(3)*jnp.sin(b)*(r1-r2)*(jnp.cos(3*a)+2*jnp.cos(b))\
          +2*jnp.sqrt(3)*u1*jnp.cos(2*a)*jnp.sin(2*b)
    return imv2
imV2x = grad(imV2, argnums = 0)
imV2y = grad(imV2, argnums = 1)

def V11(kx, ky):
    a,b = getab(kx, ky)
    v11 = e2 + (t11+3*t22)*jnp.cos(a)*jnp.cos(b) + 2*t11*jnp.cos(2*a)\
         + 4*r11*jnp.cos(3*a)*jnp.cos(b) + 2*(r11+jnp.sqrt(3)*r12)*jnp.cos(2*b)\
         + (u11+3*u22)*jnp.cos(2*a)*jnp.cos(2*b) + 2*u11*jnp.cos(4*a)
    return v11
V11x = grad(V11, argnums = 0)
V11y = grad(V11, argnums = 1)
    
def reV12(kx, ky):
    a,b = getab(kx, ky)
    rev12 = jnp.sqrt(3)*(t22-t11)*jnp.sin(a)*jnp.sin(b) + 4*r12*jnp.sin(3*a)*jnp.sin(b)\
           +jnp.sqrt(3)*(u22-u11)*jnp.sin(2*a)*jnp.sin(2*b)
    return rev12
reV12x = grad(reV12, argnums = 0)
reV12y = grad(reV12, argnums = 1)

def imV12(kx, ky):
    a,b = getab(kx, ky)
    imv12 = 4*t12*jnp.sin(a)*(jnp.cos(a)-jnp.cos(b))\
           +4*u12*jnp.sin(2*a)*(jnp.cos(2*a)-jnp.cos(2*b))
    return imv12
imV12x = grad(imV12, argnums = 0)
imV12y = grad(imV12, argnums = 1)

def V22(kx, ky):
    a,b = getab(kx, ky)
    v22 = e2 + (3*t11+t22)*jnp.cos(a)*jnp.cos(b) + 2*t22*jnp.cos(2*a)\
         +2*r11*(2*jnp.cos(3*a)*jnp.cos(b)+jnp.cos(2*b))\
         +2/jnp.sqrt(3)*r12*(4*jnp.cos(3*a)*jnp.cos(b)-jnp.cos(2*b))\
         +(3*u11+u22)*jnp.cos(2*a)*jnp.cos(2*b) + 2*u22*jnp.cos(4*a) 
    return v22
V22x = grad(V22, argnums = 0)
V22y = grad(V22, argnums = 1)

##Path in regiprocal space
#Gamma = np.array([0.,0.])
## K     = np.array([4*np.pi/3/aa,0])
## M     = np.array([np.pi/aa  ,np.pi/np.sqrt(3)/aa])
#K     = np.array([4*np.pi/3,0])                # in units of 1/a
#M     = np.array([np.pi ,np.pi/np.sqrt(3)])   # in units of 1/a
#Gammapoint = np.array([0.0, 0.0])
#Kpoint = np.array([4*np.pi/3,0])                # in units of 1/a
#Mpoint = np.array([np.pi ,np.pi/np.sqrt(3)])   # in units of 1/a
Gammapoint = np.array([0.0, 0.0]) / aa
Kpoint = np.array([4*np.pi/3,0]) / aa               # in units of 1/a
Mpoint = np.array([np.pi ,np.pi/np.sqrt(3)]) / aa  # in units of 1/a

# SOC (tab IV)
lam = 0.073 * eVtohartree
Lz  = np.array([[0, 0    ,0    ],\
                [0, 0    ,2*1.j],\
                [0,-2*1.j,0    ]])

def H0k(kk):
    htnn = np.array([[V0(*kk)                 , reV1(*kk) + 1j*imV1(*kk)  , reV2(*kk) + 1j*imV2(*kk)  ],\
                     [reV1(*kk) - 1j*imV1(*kk), V11(*kk)                  , reV12(*kk) + 1j*imV12(*kk)],\
                     [reV2(*kk) - 1j*imV2(*kk), reV12(*kk) - 1j*imV12(*kk), V22(*kk)]                 ])

    HH0= htnn
    #soc
    #H = qt.Qobj(np.kron(np.eye(2),HH0)+np.kron(qt.sigmaz(),lam/2*Lz))
    #no soc
    H = qt.Qobj(HH0)
    return H - 0.028151402249932324

def Hxk(kk):
    htnn = np.array([[V0x(*kk)                  , reV1x(*kk) + 1j*imV1x(*kk)  , reV2x(*kk) + 1j*imV2x(*kk)  ],\
                     [reV1x(*kk) - 1j*imV1x(*kk), V11x(*kk)                   , reV12x(*kk) + 1j*imV12x(*kk)],\
                     [reV2x(*kk) - 1j*imV2x(*kk), reV12x(*kk) - 1j*imV12x(*kk), V22x(*kk)]                  ])
    HH0= htnn
    #soc
    #H = qt.Qobj(np.kron(np.eye(2),HH0))
    #no soc
    H = qt.Qobj(HH0)
    return H

def Hyk(kk):
    htnn = np.array([[V0y(*kk)                  , reV1y(*kk) + 1j*imV1y(*kk)  , reV2y(*kk) + 1j*imV2y(*kk)  ],\
                     [reV1y(*kk) - 1j*imV1y(*kk), V11y(*kk)                   , reV12y(*kk) + 1j*imV12y(*kk)],\
                     [reV2y(*kk) - 1j*imV2y(*kk), reV12y(*kk) - 1j*imV12y(*kk), V22y(*kk)]                  ])
    HH0= htnn
    #soc
    #H = qt.Qobj(np.kron(np.eye(2),HH0))
    #no soc
    H = qt.Qobj(HH0)
    return H


def gammakmgammaset(nps = 50):
    #Kpoint = tmd.Kpoint
    #Mpoint = tmd.Mpoint
    #Gammapoint = tmd.Gammapoint

    GammaK = np.zeros([nps, 2])
    GammaK[:, 0] = np.linspace(Gammapoint[0], Kpoint[0], nps)
    GammaK[:, 1] = np.linspace(Gammapoint[1], Kpoint[1], nps)

    KM = np.zeros([nps, 2])
    KM[:, 0] = np.linspace(Kpoint[0], Mpoint[0], nps)
    KM[:, 1] = np.linspace(Kpoint[1], Mpoint[1], nps)

    MGamma = np.zeros([nps, 2])
    MGamma[:, 0] = np.linspace(Mpoint[0], Gammapoint[0], nps)
    MGamma[:, 1] = np.linspace(Mpoint[1], Gammapoint[1], nps)

    GammaKset = Kpointset(kpoints = GammaK,
                          line = True,
                          startpoint = '$\Gamma$', endpoint = 'K')
    KMset = Kpointset(kpoints = KM,
                      line = True, 
                      coordorigin = GammaKset.coord[-1],
                      startpoint = 'K', endpoint = 'M')
    MGammaset = Kpointset(kpoints = MGamma,
                          line = True,
                          coordorigin = KMset.coord[-1],
                          startpoint = 'M', endpoint = '$\Gamma$')

    kset = GammaKset
    kset.join(KMset)
    kset.join(MGammaset)

    return kset
